/*
  * Copyright (c) 2019 Ignacio Francisco Ramírez Paulino and Ignacio Hounie
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Affero General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or (at
  * your option) any later version.
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero
  * General Public License for more details.
  * You should have received a copy of the GNU Affero General Public License
  *  along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * \file paco_grid_mapping.c
 * \brief implementation of the grid patch mapping scheme
 */
#include <math.h>
#include <omp.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "paco_log.h"
#include "paco_types.h"

static index_t input_height; ///< image nrows
static index_t input_width; ///< image ncols
static index_t num_input_samples; ///< total number of samples = num_input_pixels
static index_t patch_width; ///< patch width. Provided
static index_t patch_height; ///< patch width. Provided
static index_t num_patch_samples; ///< 3D: width x height
static index_t *rel_idx; ///< relative linear indexes of pixels within a patch, computed for given patch dimensions.
static index_t  num_mapped_patches; ///< number of mapped patches
static index_t *idx_mapped_patches; ///< linear patch index of mapped patches (left, upper corner)
static index_t  num_mapped_samples; ///< number of pixels which are mapped to patch space
static index_t* idx_mapped_samples; ///< linear indexes of pixels which are mapped to patch space
static sample_t* fact_mapped_samples; ///< normalization factor for mapped pixels, used for stitching

//============================================================================

index_t get_patch_width() {
    return patch_width;
}

//============================================================================

index_t get_patch_height() {
    return patch_height;
}

//============================================================================

index_t get_num_patch_samples() {
    return num_patch_samples;
}

//============================================================================

index_t get_input_height() {
    return input_height;
}

//============================================================================

index_t get_input_width() {
    return input_width;
}

//============================================================================

index_t get_num_input_samples() {
    return input_width*input_height;
}

//============================================================================

index_t get_num_mapped_samples() {
    return num_mapped_samples;
}

//============================================================================

index_t get_num_mapped_patches() {
    return num_mapped_patches;
}

//============================================================================

const index_t* get_idx_mapped_samples() {
    return idx_mapped_samples;
}

//============================================================================

void paco_mapping_destroy() {
    // delete stuff...
    free(rel_idx);
    rel_idx = 0;
    free(idx_mapped_patches);
    idx_mapped_patches = 0;
    free(idx_mapped_samples);
    idx_mapped_samples = 0;
    free(fact_mapped_samples);
    fact_mapped_samples = 0;
}


//============================================================================

void extract_patch(sample_t* patch, const gsl_matrix_float* samples, const index_t i, const index_t j) {
    const index_t ph = patch_height;
    const index_t pw = patch_width;
    assert(i <= samples->size1-ph);
    assert(j <= samples->size2-pw);
    const float* offset = gsl_matrix_float_const_ptr(samples,i,j);
    const index_t m = num_patch_samples;
    for (index_t k = 0; k < m; ++k) {
        patch[k] = offset[rel_idx[k]];
    }
}

//============================================================================
char is_patch_complete(const gsl_matrix_float* mask, const index_t i0, const index_t j0) {
    const float* offset = gsl_matrix_float_const_ptr(mask,i0,j0);
    const index_t m = num_patch_samples;
    for (index_t k = 0; k < m; ++k) {
        if (offset[rel_idx[k]] != 0) {
            return 0;
        }
    }
    return 1;
}

//============================================================================

void extract_mapped_patch(sample_t* patch, const gsl_matrix_float* samples, const index_t j) {
    assert(j < num_mapped_patches);
    index_t offset = idx_mapped_patches[j];
    const index_t m = num_patch_samples;
    for (index_t i = 0; i < m; ++i) {
        patch[i] = samples->data[offset+rel_idx[i]];
    }
}

//
//============================================================================
//

void extract_patches(gsl_matrix_float *X, const gsl_matrix_float* I) {

    const index_t m = num_patch_samples;
    const index_t n = num_mapped_patches;
    const index_t* poff = idx_mapped_patches;
    const sample_t* pI = I->data;
    const index_t* pri = rel_idx;

    for (int j = 0; j < n; j ++) {
        sample_t* pXj = &X->data[j*X->tda];
        const index_t off = poff[j];
        for (int i= 0; i < m; i ++) {
            *(pXj++) = pI[off + pri[i]];
        }
    }
}

//
//============================================================================
//
void stitch_patches(gsl_matrix_float* I, const gsl_matrix_float* X) {
    //
    // zero out mapped pixels
    //
    sample_t* pI = I->data;
    const sample_t* pX = X->data;
    const index_t nmapped = num_mapped_samples;
    const index_t n = num_mapped_patches;
    const index_t m = num_patch_samples;
    const index_t* pimpi = idx_mapped_samples;
    const index_t* pimpa = idx_mapped_patches;
    //
    // initialize to zero
    //
    for (int k = 0; k < nmapped; k++)  {
        pI[ pimpi[k] ] = 0;
    }
    //
    // add up all the patch estimates (from X) into their corresponding places in I
    //
    const index_t* pri = rel_idx;
    for (int j = 0; j < n; j ++) {
        const sample_t* pXj = &pX[X->tda * j];
        sample_t* pP = pI + pimpa[j];
        for (int i = 0; i < m; i ++) {
            pP[ pri[i] ] += pXj[ i ];
        }
    }
    //
    // Normalize
    //
    const sample_t* pf = fact_mapped_samples;
    for (int k = 0; k < nmapped; k++)  {
        pI[ pimpi[k] ] *= pf[ k ];
    }
}

/**
 * ============================================================================
 *
 * This function creates an instance of a a mapping between an image
 * and a set of patches; it implements ALGORITHM 5 in the IPOL paper.
 *
 * If a mask is provided, then only those patches which contain ones in the mask are extracted.
 * If no mask is provided, then all patches will be extracted.
 *
 * In any case, a number of index sets are precomputed here to speed
 * up the extraction and stitching operations. Namely:
 *
 * - relative linear indexes of a patch w.r.t. the upper-left corner; this
 * enables the use of only one loop for any N-dimensional problem
 *
 * - linear indexes of the upper-left corners of all the mapped patches.
 *
 * - linear indexes of the pixels which are mapped to one (or more) patches.
 *
 * - normalization factors for the mapped pixels. This allows for a very
 *   efficient stitching operation.
 */
void paco_grid_mapping_init(const gsl_matrix_float* input, const gsl_matrix_float* mask, index_t w, index_t s) {
    paco_info("Creating patch grid mapping ...\n");

    assert(w > 0);
    assert(s > 0);
    //
    // store basic information such as the signal dimensions
    // the number of channels, etc.
    //
    input_height     = mask->size1;
    input_width      = mask->size2;
    num_input_samples = input_height * input_width;
    patch_width = w;
    patch_height = w;
    num_patch_samples  = w * w;
    //
    // compute the grid size in both directions.
    // if a mask is provided, not all points of the grid are mapped,
    // but only those for which the corresponding patches starting at
    // them contain at least missing sample.
    //
    const index_t ni   = (input_height - w  + 1)/s;
    const index_t nj   = (input_width - w + 1)/s;
    const index_t num_patches  = nj * ni;
    paco_debug("input image size %d x %d\n", input_height, input_width);
    paco_debug("grid size %d x %d = %d\n", ni, nj, num_patches);
    paco_debug("patch size %d x %d = %d \n", w, w, num_patch_samples);
    //
    //  linear pixel indexes relative to a patch
    //
    rel_idx = (index_t*)malloc(num_patch_samples*sizeof(index_t));
    //
    // gsl matrices of size mxn are contiguous in memory, each row taking tda = m*sizeof(double)
    // we store M x N  images as M * N matrices
    //
    for (int i = 0, k = 0; i < w; i++) {
        for (int j= 0; j < w; j++) {
            rel_idx[k++] = i*input->tda + j;
        }
    }
    if (mask == NULL) {
        //
        // no mask: extract all patches from image
        // store patch linear index (leftmost & uppermost patch pixel)
        //
        num_mapped_patches = ni * nj;
        idx_mapped_patches = (index_t*)malloc(num_mapped_patches*sizeof(index_t));
        int k = 0;
        for (int i = 0; i <= input_height-w; i += s) {
            for (int j= 0; j <= input_width-w; j += s, k++) {
                idx_mapped_patches[k] = input->tda*i + j;
            }
        }
        assert(k == num_mapped_patches);
        //
        // if there is no mask, then all pixels are mapped
        //
        num_mapped_samples = num_input_samples;
        idx_mapped_samples = 0;
    } else {
        //
        // mask provided
        //
        // hits_map is an auxiliary array which, in the end, counts how many times any
        // given pixel appears in the patch matrix.
        //
        // this is later used to compute the normalization factors for stitching,
        // and also used for determining which patches contain missing pixels.
        //
        gsl_matrix_float* hits_map = gsl_matrix_float_alloc(input_height, input_width);
        gsl_matrix_float_set_zero(hits_map);
        //
        // count incomplete patches
        //
        int ninc = 0;
        for (int i = 0; i <= input_height-w; i += s) {
            for (int j= 0; j <= input_width-w; j += s) {
                if ( !is_patch_complete(mask,i,j) ) {
                    ninc++;
                }
            }
        }

        num_mapped_patches = ninc;
        paco_debug("incomplete patches=%d\n",ninc);

        // store incomplete patch indexes
        idx_mapped_patches = (index_t*)malloc(ninc * sizeof(index_t));

        ninc = 0;
        for (int i = 0; i <= input_height-w; i += s) {
            for (int j= 0; j <= input_width-w; j += s) {
                if ( !is_patch_complete(mask,i,j) ) {
                    idx_mapped_patches[ninc++] = input->tda*i+j;
                    for (int ii = 0; ii < w; ii ++) {
                        for (int jj= 0; jj < w; jj++) {
                            // increment hits count
                            (*gsl_matrix_float_ptr(hits_map,i+ii,j+jj)) ++;
                        }
                    }
                }
            }
        }
        paco_debug("incomplete patches=%d\n",ninc);
        //
        // Count mapped pixels = any pixel for which hits_map is > 0
        //
        int nmapped = 0;
        for (int i = 0, k = 0; i < input_height; i ++) {
            for (int j= 0; j < input_width; j ++, k++) {
                if (gsl_matrix_float_get(hits_map,i,j))  {
                    nmapped++;
                }
            }
        }
        idx_mapped_samples  = (index_t*)malloc(nmapped * sizeof(index_t));
        paco_debug("mapped pixels=%d\n",nmapped);
        num_mapped_samples  = nmapped;
        //
        // Store mapped and projected linear indexes
        //
        fact_mapped_samples = (sample_t*) calloc(num_mapped_samples,sizeof(sample_t));
        //
        // record mapped pixel indexes
        // compute correspondining normalization factors from hits_map
        //
        nmapped = 0;
        for (int i = 0, k = 0; i < input_height; i ++) {
            for (int j= 0; j < input_width; j ++, k++) {
                double hits = gsl_matrix_float_get(hits_map,i,j);
                if (hits > 0)  {
                    idx_mapped_samples[nmapped] = input->tda*i+j;
                    fact_mapped_samples[nmapped] = 1.0 / (double) hits;
                    nmapped++;
                }
            }
        }
        //if (level_debug()) {
        //    paco_image_st* hits_img = paco_image_from_samples(hits_map,COLORSPACE_GRAY);
        //    write_png_file("hits_map.png",hits_img);
        //    paco_image_free(hits_img);
        //}
        gsl_matrix_float_free(hits_map);
    }
}
